% DEMODCT3_JPX_Lap.M
% Perform DCT analysis of Lenna using two levels of 4x4 DCTs and
% calculate entropies etc.
%
% The '_Lap' version includes the overlapped filtering option which precedes each DCT stage.
% enable_lap controls the lapped filtering: 0 disables it; 1 enables it at
% level 1; 2 enables it at level 2; and 3 enables it at both levels.
%
% Nick Kingsbury, Cambridge University, October 2011.

% clfigs
global printon
printon = input('Write print files (0=no, 1=yes, 2=GIF):');
% printon = 0;
enable_lap = input('Enable overlapped filters (0=none, 1=level 1, 2=level 2, 3 = both):');
qscale = input('Quantiser scale factor (1=normal ie 15 and Qlum, 2=twice normal');

% 1-D DCT matrix.
T = dctmat(4);

% 1-D 4-pt overlap transform matrix pot4 and its inverse ipot4.
% NOTE: o2 contains a _conventional_ pi/8 rotate with 2 positive cosine
% terms, rather than the rotate-with-flip which is used elsewhere.
o1 = sqrt(0.5) * [1 0 0 1; 0 1 1 0; 0 1 -1 0; 1 0 0 -1];
c = cos(pi/8); s = sin(pi/8);
o2 = [1 0 0 0; 0 1 0 0; 0 0 c s; 0 0 -s c];
sc = 0.8272; % Scale factor
% sc = 1;
pot4 = o1 * diag([sc sc 1/sc 1/sc]) * o2 * o1;  % Overlap filter
ipot4 = o1 * o2' * diag([1/sc 1/sc sc sc]) * o1; % Inverse of overlap filter

load lenna
X = X - 120;
Xorig = X;

% Perform overlap prefiltering at level 1.
if (enable_lap == 1) | (enable_lap == 3),
    sx = size(X);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Xe = X(t1,t2); % Symmetrically extend X by 2 rows/columns at each edge.
    Xe = colxfm(colxfm(Xe,pot4)',pot4)';
    X = Xe(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

disp('Doing level-1 DCT'); X = colxfm(colxfm(X,T)',T)'; 
X = regroup(X,4);

% Level 2 of the DCT.
X2 = X(1:(end/4),1:(end/4));  %Lo-Lo band subimage.

% Perform overlap prefiltering at level 2.
if (enable_lap >= 2),
    sx = size(X2);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Xe = X2(t1,t2); % Symmetrically extend X by 2 rows/columns at each edge.
    Xe = colxfm(colxfm(Xe,pot4)',pot4)';
    X2 = Xe(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

disp('Doing level-2 DCT'); X2 = colxfm(colxfm(X2,T)',T)'; 
X2 = regroup(X2,4);

Xa = X;
Xa(1:(end/4),1:(end/4)) = X2;

% Fig 3.2c: DCT reordered into subimages grouped by coefficient type.
figure;
draw(min(max(Xa,-400),400))
drawnow
if printon==2,
  drawfile(X,'i8dct31xr.gif');
elseif printon==1,
  print -deps i8dct31xr.eps
end

% Plot the 16 entropies of X using mesh and plot.
ent4 = dctent(X,qscale*15,4);
ent4b = dctent(X2,qscale*15,4);

mat2scr(ent4,'%8.4f','Level 1, 4 x 4 DCT entropy matrix');
mat2scr(mean(ent4(:)),'Mean entropy of level 1 = %.4f bits');

mat2scr(ent4b,'%8.4f','Level 2, 4 x 4 DCT entropy matrix');
mat2scr(mean(ent4b(:)),'Mean entropy of level 2 = %.4f bits');

ent44 = ent4;
ent44(1,1) = mean(ent4b(:));
ent44mean = mean(ent44(:));

mat2scr(ent44,'%8.4f','Overall 4 x 4 DCT entropy matrix');
mat2scr(ent44mean,'Mean entropy of both levels = %.4f bits');

figure; setfig(gcf);
subplot(1,2,1);
mesh(ent4');
view(110,20);
axis('square');
axis([0 5 0 5 0 8]);
grid on
title('Mesh plot of DCT entropy','FontSize',16);

subplot(1,2,2);
plot(ent4');
axis('square');
axis([0 4 0 8]);
grid on
title('Row plots of DCT entropy','FontSize',16);
xlabel('column','FontSize',16);
ylabel('Entropy (bit/pel)','FontSize',16);
text(1,ent4(1,1),'row 1','horiz','center','vert','bot');
text(1,ent4(2,1),'2 ','horiz','right','vert','mid');
text(1,ent4(3,1),'3 ','horiz','right','vert','mid');
text(1,ent4(4,1),'4 ','horiz','right','vert','mid');
drawnow

% Quantise using qstep=15.
qstep = qscale*15;
Y = dctquant(Xa,qstep,4);

Y2 = Xa(1:(end/4),1:(end/4));
Y2 = regroup(Y2,16);
disp('Doing Inv-DCT'); Y2 = colxfm(colxfm(Y2',T')',T');

% Perform overlap postfiltering at level 2.
if (enable_lap >= 2),
    sx = size(Y2);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Ye = Y2(t1,t2); % Symmetrically extend Y by 2 rows/columns at each edge.
    Ye = colxfm(colxfm(Ye',ipot4)',ipot4);
    Y2 = Ye(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

Y(1:(end/4),1:(end/4)) = Y2;

% Perform inverse transform.
Y = regroup(Y,64);
disp('Doing Inv-DCT'); Y = colxfm(colxfm(Y',T')',T');

% Perform overlap postfiltering at level 1.
if (enable_lap == 1) | (enable_lap == 3),
    sx = size(Y);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Ye = Y(t1,t2); % Symmetrically extend Y by 2 rows/columns at each edge.
    Ye = colxfm(colxfm(Ye',ipot4)',ipot4);
    Y = Ye(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

rmseq15 = rms(Y(:)-Xorig(:))

figure; setfig(gcf);
subplot(1,2,1);
mesh(ent44');
view(110,20);
axis('square');
axis([0 5 0 5 0 8]);
grid on
title('Mesh plot of 2-level DCT entropy','FontSize',16);

subplot(1,2,2);
plot(ent44');
axis('square');
axis([0 4 0 8]);
grid on
title('Row plots of 2-level DCT entropy','FontSize',16);
xlabel('column','FontSize',16);
ylabel('Entropy (bit/pel)','FontSize',16);
text(1,ent44(1,1),'row 1','horiz','center','vert','bot');
text(1,ent44(2,1),'2 ','horiz','right','vert','mid');
text(1,ent44(3,1),'3 ','horiz','right','vert','mid');
text(1,ent44(4,1),'4 ','horiz','right','vert','mid');
% text(2,6.5,sprintf('Mean entr=%.4f',ent44mean),'vert','mid','horiz','cent');
% text(2,5.5,sprintf('RMSE=%.4f',rmseq15),'vert','mid','horiz','cent');
drawnow
if printon==1,
  print -deps i8dct32xr.eps
end


figure;
draw(max(min(Y,120),-120),[],[4 184]);
set(gcf,'numbertitle','off','name',sprintf('4-point DCT on 2 levels: Qstep = %.1f',qstep));
drawnow

% Now experiment with approximation to JPEG Qlum matrix
qlum = jpegqtab;
t = 1:2:7;
qlum4 = round((qlum(t,t) + qlum(t+1,t) + qlum(t,t+1) + qlum(t+1,t+1))/4);
qlum4(1,1) = 15;  % Continue to use qstep = 15 for level 2.
% qlum4
qlum4 = [
    15    15    25    57
    15    23    53    90
    25    53    91   100
    57    90   100   106];
% qlum4 = [15 15 25 45
%          15 25 45 60
%          25 45 60 75
%          45 60 75 75];
qlum4

% Quantise using qstep=qlum for level 1 and qstep=15 for level 2.
% qstep = qlum;
[Y,ent4qlum,rmsec] = dctquant(Xa,qscale*qlum4,4);
[Y2,ent4qlum2,rmsec2] = dctquant(X2,qscale*qlum4(1,1),4);
ent4qlum,rmsec
ent4qlum2,rmsec2
ent4qlum(1,1) = mean(ent4qlum2(:));
ent4qlmean = mean(ent4qlum(:))
rmseqlmean = rms(rmsec(:))

Y2 = regroup(Y2,16);
disp('Doing Inv-DCT'); Y2 = colxfm(colxfm(Y2',T')',T');

% Perform overlap postfiltering at level 2.
if (enable_lap >= 2),
    sx = size(Y2);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Ye = Y2(t1,t2); % Symmetrically extend Y by 2 rows/columns at each edge.
    Ye = colxfm(colxfm(Ye',ipot4)',ipot4);
    Y2 = Ye(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

Y(1:(end/4),1:(end/4)) = Y2;

% Perform inverse transform.
Y = regroup(Y,64);
disp('Doing Inv-DCT'); Y = colxfm(colxfm(Y',T')',T');

% Perform overlap postfiltering at level 1.
if (enable_lap == 1) | (enable_lap == 3),
    sx = size(Y);
    t1 = [2 1 1:sx(1) sx(1) sx(1)-1];
    t2 = [2 1 1:sx(2) sx(2) sx(2)-1];
    Ye = Y(t1,t2); % Symmetrically extend Y by 2 rows/columns at each edge.
    Ye = colxfm(colxfm(Ye',ipot4)',ipot4);
    Y = Ye(3:end-2,3:end-2); % Remove 2 rows/columns at each edge.
end

rmseql = rms(Y(:)-Xorig(:))

figure;
draw(max(min(Y,120),-120),[],[4 184]);
set(gcf,'numbertitle','off','name',sprintf('4-point DCT on 2 levels: Qstep = %.1f * JPEG lum.',qscale));
drawnow
if printon==2,
  drawfile(Y,'i8dct33xr.gif');
elseif printon==1,
  print -deps i8dct33xr.eps
end


figure; setfig(gcf);
subplot(1,2,1);
mesh(ent4qlum');
view(110,20);
axis('square');
axis([0 5 0 5 0 8]);
grid on
title('Mesh plot of 2-level DCT entropy with Qlum','FontSize',16);

subplot(1,2,2);
plot(ent4qlum');
axis('square');
axis([0 4 0 8]);
grid on
title('Row plots of 2-level DCT entropy with Qlum','FontSize',16);
xlabel('column','FontSize',16);
ylabel('Entropy (bit/pel)','FontSize',16);
text(1,ent4qlum(1,1),'row 1','horiz','center','vert','bot');
text(1,ent4qlum(2,1),'2 ','horiz','right','vert','mid');
text(1,ent4qlum(3,1),'3 ','horiz','right','vert','mid');
text(1,ent4qlum(4,1),'4 ','horiz','right','vert','mid');
text(2,6.5,sprintf('Mean entr=%.4f',ent4qlmean),'vert','mid','horiz','cent');
text(2,5.5,sprintf('RMSE=%.4f',rmseql),'vert','mid','horiz','cent');
drawnow

if (enable_lap >= 1),
    save ent44 ent44 ent4b rmseq15 ent4qlum rmseql
else
    save ent44nofilt ent44 ent4b rmseq15 ent4qlum rmseql
end



