function [bits, huffval, huffent] = genhuff2(huffhist)

% function [bits, huffval, huffent] = genhuff2(huffhist)
% Generate the JPEG tables bits and huffval from the 
% histogram of values huffhist (must be < 256 values).
% This is based on the algorithms in JPEG Appendix K.2.
% This version allows an arbitrary histogram size and does not
% prohibit the all 1's code word, but it does restrict the max
% word length to be 16 bits.
% huffent is the vector of entropies for each entry in huffhist,
% assuming the integer Huffman word lengths instead of -log2(p).

scale = (255*256) / sum(huffhist);

% Scale huffhist to sum to (64K - 256).
freq = huffhist(:)*scale;

% Add 1 to all valid points so they are given a code word.
% With the scaling, this should ensure no code words exceed 16
% bits and is probably a better method with poor statistics
% than the JPEG method of fig K.3.
freq = freq + 2;  

% Reshape to a vector and add 257th point to reserve the FFFF
% codeword.
% Also add a small negative ramp so that min() always picks the
% larger index when 2 points have the same probability.
freq = [freq(:); 2] - [0:length(freq)]'*1e-6;

codesize = zeros(size(freq));
others = -ones(size(freq));

% Find Huffman code sizes: JPEG fig K.1, procedure Code_size

% Find non-zero entries in freq and loop until 1 entry left.
nz = find(freq > 0);
while length(nz) > 1,

% Find v1 for least value of freq(v1) > 0.
  [y,i] = min(freq(nz));
  v1 = nz(i);

% Find v2 for next least value of freq(v2) > 0.
  nz = nz([1:(i-1)  (i+1):length(nz)]);  % Remove v1 from nz.
  [y,i] = min(freq(nz));
  v2 = nz(i);

% Combine frequency values.
  freq(v1) = freq(v1) + freq(v2);
  freq(v2) = 0;

  codesize(v1) = codesize(v1) + 1;
  while others(v1) > -1,
    v1 = others(v1);
    codesize(v1) = codesize(v1) + 1;
  end
  others(v1) = v2;
  codesize(v2) = codesize(v2) + 1;
  while others(v2) > -1,
    v2 = others(v2);
    codesize(v2) = codesize(v2) + 1;
  end
  nz = find(freq > 0);
end

% Generate the Huffman entropy.
huffent = (huffhist(:) / sum(huffhist)) .* codesize(1:prod(size(huffhist)));

% Find no. of codes of each size: JPEG fig K.2, procedure Count_BITS
maxcode = max(codesize);
bits = zeros(max(16, maxcode),1);
for i = 1:(length(codesize)-1),
  if codesize(i) > 0, 
    bits(codesize(i)) = bits(codesize(i)) + 1;
  end
end

% Code length limiting not needed since 1's added earlier to all valid
% codes.
if maxcode > 16,
  error('GENHUFF.M: max(codesize) > 16');
end

% Sort codesize values into ascending order and generate huffval:
% JPEG fig K.4, procedure Sort_input.

huffval = [];
t = 1:(length(codesize)-1);
for i = 1:maxcode,
  ii = find(codesize(t) == i);
  huffval = [huffval; ii];
end

% Correct for indices of codesize going 1:256 instead of 0:255.
huffval = huffval - 1;

if length(huffval) ~= sum(bits),
  disp('Warning! GENHUFF.M: length of huffval ~= sum(bits)');
end
return
